------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------
--
-- This file contains all stored procedures related to inserting the model's data into database 
--
--
------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------

-- search for extensions in l3d schema
SET search_path TO l3d;

-- function : l3d.sp_exists_temporary_table(varchar)

drop table if exists l3d.ins_stats ;

CREATE TABLE l3d.ins_stats (
	cache_id int8 NULL,
	table_name varchar NULL,
	time_start timestamp NULL,
	time_end timestamp NULL,
	records_count int4 NULL,
	millis_collect int4 NULL,
	millis_flush int4 NULL,
	records_commited int4 NULL
);
CREATE UNIQUE INDEX idx_ins_stats ON ins_stats USING btree (cache_id, table_name);

alter table l3d.ins_stats owner to i3dhub; 



drop function if exists l3d.sp_exists_temporary_table(varchar);

create or replace function l3d.sp_exists_temporary_table(p_table_name varchar)
/**
 *  select l3d.sp_exists_temporary_table('tmp_insert_node')
 */
returns bool
LANGUAGE plpgsql
AS $function$
begin	
	return exists(select * from pg_catalog.pg_tables t, pg_namespace ns where t.tablename = p_table_name and t.schemaname = ns.nspname and oid = pg_my_temp_schema());	
end;
$function$;

ALTER FUNCTION l3d.sp_exists_temporary_table(varchar) OWNER TO i3dhub;

-- function: l3d.sp_insert_cache(l3d.cache.uuid%type, l3d.cache.ext_cache_id0%type, l3d.cache.ext_cache_id1%type)

drop function if exists  l3d.sp_insert_cache(l3d.cache.uuid%type, l3d.cache.ext_cache_id0%type, l3d.cache.ext_cache_id1%type);

CREATE OR REPLACE FUNCTION l3d.sp_insert_cache(p_uuid l3d.cache.uuid%type default null, p_cacheid0 l3d.cache.ext_cache_id0%type default null, p_cacheid1 l3d.cache.ext_cache_id1%type default null)
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 * 	  // Inserting cache, which their cacheid0 and cacheid1 are provided
 *    select * from l3d.sp_insert_cache('5b2f17fa79e7cdc3-45d38cf7-4645-b994', 1000, 1001) as a 
 */
	db_id l3d.cache.id%type := null;
	r_cache record;
BEGIN 		

	for r_cache in insert into l3d.cache (ext_cache_id0, ext_cache_id1, uuid) select p_cacheid0, p_cacheid1, p_uuid returning *
	loop
		return r_cache.id;
	end loop;

end
$function$;

ALTER FUNCTION l3d.sp_insert_cache(l3d.cache.uuid%type, l3d.cache.ext_cache_id0%type, l3d.cache.ext_cache_id1%type) OWNER TO i3dhub;


-- function: l3d.sp_insert_fetch_cache(integer)

drop function if exists l3d.sp_insert_fetch_cache(integer);

CREATE OR REPLACE FUNCTION l3d.sp_insert_fetch_cache(p_id integer DEFAULT NULL::integer)
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 * 	  // Fetching cache from database as initial step for starting data insertion of its data
 *    select * from l3d.sp_insert_fetch_cache(1) as a 
 */
	db_id l3d.cache.id%type := null;

BEGIN 		
	
	DISCARD temp;
	
	if not sp_exists_temporary_table('tmp_insert_cache') then
		create temp table tmp_insert_cache as select * from l3d.cache where id = p_id;	
	else
		truncate table tmp_insert_cache;
		insert into tmp_insert_cache select * from l3d.cache where id = p_id;		
	end if;

--	perform l3d.sp_insert_clear();

	return p_id;
end
$function$;

ALTER FUNCTION l3d.sp_insert_fetch_cache(integer) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_stats(varchar,integer,integer);


CREATE OR REPLACE FUNCTION l3d.sp_insert_stats(p_table_name varchar, p_row_count integer default null, p_threshold integer default 4000)
 RETURNS bigint
 LANGUAGE plpgsql
AS $function$
declare 
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer := p_row_count;
begin 
	if v_row_count is null then 
	
		get diagnostics v_row_count = row_count;		
		
	end if;
	
	if not exists(select * from l3d.ins_stats where cache_id = v_cache_id and table_name = p_table_name) then 
	
		insert into l3d.ins_stats(cache_id, table_name, time_start, records_count, records_commited) select v_cache_id, p_table_name, current_timestamp, v_row_count, 0;			
	else

		update l3d.ins_stats set records_count = coalesce( records_count, 0 )  + v_row_count where cache_id = v_cache_id and table_name = p_table_name;
	
	end if;		

	select sum( coalesce( records_count, 0 ) - coalesce( records_commited, 0 ) ) into v_row_count from l3d.ins_stats where cache_id = v_cache_id;

	if v_row_count >= p_threshold then
		perform l3d.sp_insert_flush(false);		
	end if;

	return 0;

end;	
$function$;

ALTER FUNCTION l3d.sp_insert_stats(varchar,integer,integer) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_nodes(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_nodes(p_nodes_data character varying[] DEFAULT NULL::character varying[])
 RETURNS bigint
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting Nodes linking them to last cache inserted
 *			select * from l3d.sp_insert_nodes(ARRAY['[0]-,-[Node 1]','[1]-,-[Node 2]','[2]-,-[Node 3]']) as a   
 */
	v_table_exists bool := sp_exists_temporary_table('tmp_ins_node');	     
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result bigint;
BEGIN 			
	
	if p_nodes_data is null  then
	
			if v_table_exists then	
			
				if exists(select * from tmp_ins_node)   then			
				
					with ins_node as
					(
						insert into l3d.node(ext_node_id, label, parent_node_id , cache_id) 
						select ext_node_id, label, parent_node_id, v_cache_id as cache_id	
						from
						(
							select * from tmp_ins_node order by ext_node_id						
						) a
						returning id, ext_node_id				
					) , upd_stats as 
					(
						update l3d.ins_stats s set time_end = current_timestamp, records_commited = records_commited  + (select count(*) from ins_node)
						where cache_id = v_cache_id and table_name = 'node'						
					)
					insert into tmp_insert_node_range(min_id, max_id, offset_id)
					select min(ext_node_id), max(ext_node_id), (id - ext_node_id)
					from ins_node i
					group by (id - ext_node_id);
																
					truncate table tmp_ins_node;			
				
				end if;				
			
			end if;
		
	elsif array_length(p_nodes_data, 1) > 0 then
					
		if not v_table_exists then											
				
			create temp table tmp_ins_node(ext_node_id int, label varchar, parent_node_id int);
			create unique index idx_tmp_ins_node on tmp_ins_node(ext_node_id);		

			create temp table tmp_insert_node_range(min_id integer, max_id integer, offset_id bigint);
			create unique index idx_tmp_insert_node_range on tmp_insert_node_range(min_id, max_id);		
		
		end if;
	
		insert into tmp_ins_node(ext_node_id, label, parent_node_id) 					
		select rd[1]::integer  as ext_node_id,  rd[2] as label,  rd[3]::integer as parent_node_id
		from 
		( 			
			select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd  from unnest(p_nodes_data) as rd
			
		) a;
		
		select max(ext_node_id) into v_result from tmp_ins_node;
	
		perform l3d.sp_insert_stats('node', array_length(p_nodes_data, 1));
		
	end if;

	return v_result;				
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_nodes Message => %\r\n\r\nContext : %\r\n\r\nData : % ' , v_err_txt ,  v_context, p_nodes_data ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_nodes(varchar[]) OWNER TO i3dhub;

drop FUNCTION if exists l3d.sp_insert_metadata(varchar[],varchar);

CREATE OR REPLACE FUNCTION l3d.sp_insert_metadata(p_metadata varchar[] DEFAULT NULL::varchar[], p_value_type varchar DEFAULT 's'::varchar)
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *    // Inserting metadata, with string values, each triplet corresponds to [key]-,-[value]-,-[ext_node_id]
 *    select * from l3d.sp_insert_metadata(array['[Name]-,-[PartBody]-,-[0]','[Source]-,-[as1.CATProduct]-,-[1]'],'s') as a
 
 *    // Inserting metadata, with numeric values, each triplet corresponds to [key]-,-[value]-,-[ext_node_id]
 *    select * from l3d.sp_insert_metadata(array['[LAYER]-,-[260]-,-[2]','[Sachnummer]-,-[9159860]-,-[2]'],'n') as a
 * 	 select l3d.sp_insert_metadata();
 * *     
*/     
	v_table_exists bool := sp_exists_temporary_table('tmp_ins_metadata');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
BEGIN 		
	if p_metadata is null then
	
		if v_table_exists  then 
			if exists(select * from tmp_ins_metadata) then

				insert into l3d.metadata(key, value_str,value_num, node_id) 
				select key, value_str,value_num, mt.ext_node_id + rng.offset_id 
				from 
				(
					select	key, 
							value_str,
							value_num, 
							unnest(nodes_ids) as ext_node_id  
					from  tmp_ins_metadata
				)mt, tmp_insert_node_range rng
				where mt.ext_node_id between rng.min_id and rng.max_id;			
							
				get diagnostics v_row_count = ROW_COUNT;			
									
				update l3d.ins_stats s set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'metadata';						
								
				truncate table tmp_ins_metadata;
			
			end if;		
		end if;	
	
	elsif array_length(p_metadata, 1) > 0 then
	
		if not v_table_exists then		
			create temp table tmp_ins_metadata(key varchar, value_str varchar, value_num numeric, nodes_ids integer[]);				
		end if;

	
		insert into tmp_ins_metadata(key, value_str,value_num, nodes_ids)
		select key, case when not is_numeric then value end, case when is_numeric then replace(value, ',', '.')::numeric end, nodes_ids
		from
		(
			select 	key, value, array_agg( node_id ) as nodes_ids, a.value_type = 'n' as is_numeric
			from
			(
				select coalesce( substring( rd_array[ rd_length - 3 ] , '(s|n)' ), p_value_type ) as value_type, rd_array[ rd_length - 2 ] as key, rd_array[ rd_length - 1 ] as value, rd_array[ rd_length ]::integer as node_id
				from
				(
					select rd_array, array_length( rd_array, 1 ) as rd_length
					from
					(
						select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_metadata ) as rd
					) a				
				) as a				
			) as a
			group by a.key, a.value, a.value_type
		) a;		

		get diagnostics v_row_count = ROW_COUNT;	
	
	
		perform l3d.sp_insert_stats('metadata', v_row_count);											
	end if;	
	
	return 0;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_metadata Message => %\r\n\r\nContext : %\r\n\r\nData : % ' , v_err_txt ,  v_context, p_metadata ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_metadata(varchar[],varchar) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_tags(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_tags(p_tags varchar[] DEFAULT NULL::varchar[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:  Geometrical Set.1
 *    // Inserting tags, each triplet corresponds to [name]-,-[value]-,-[ext_node_id]
 *    select * from l3d.sp_insert_tags(array['[label]-,-[8185432_SOL1]-,-[0]','[label]-,-[8185432_SOL2]-,-[1]', '[label]-,-[Geometrical Set.1]-,-[2]']) as a
 * *     
*/     
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_tags_values');	
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
begin	
	if  p_tags is null then	
	
		if v_exist_table then 
			if exists(select * from tmp_ins_tags_values) then	
						
			
				insert into tags_values (node_id, value, tag_name_id) 
				select tv.ext_node_id + rng.offset_id, tv.value, tv.name_id
				from 
					tmp_ins_tags_values tv, tmp_insert_node_range rng
				where  
					tv.ext_node_id between rng.min_id and rng.max_id;	
			
				get diagnostics v_row_count = ROW_COUNT;
							
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'tags';		
			
				truncate table tmp_ins_tags_values;
			
			end if;			
		end if;	
	
	elsif array_length(p_tags, 1) > 0 then
	
		if not v_exist_table then 
				
			create temp table tmp_ins_tags_values(name varchar, value varchar, ext_node_id integer, name_id integer);
			create index idx_tmp_ins_tags_values on tmp_ins_tags_values(name_id);			
							
		end if;
				
	
		insert into tmp_ins_tags_values(name, value, ext_node_id, name_id)
		select a.name, a.value, a.ext_node_id, tn.id
		from
		(
			select rd_array[ 1 ] as name, rd_array[ 2 ] as value, substring( rd_array[ 3 ] , '[0-9]+')::integer as ext_node_id					
			from 
			(
				select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_tags ) as rd				
			) as a											
		) as a
		left join l3d.tag_name tn on tn."name" = a.name;
	
		if exists(select * from tmp_ins_tags_values where name_id is null) then 
			
			with ins_tag_name as
			(
				insert into l3d.tag_name (name) select distinct name from tmp_ins_tags_values where name_id is null
				returning id, name				
			)
			update tmp_ins_tags_values v set name_id = n.id from ins_tag_name n where v.name = n.name and v.name_id is null;					
		
		end if;
	
		perform l3d.sp_insert_stats('tags', array_length(p_tags, 1));	
					
	end if;
	
	return 0;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_tags Message => %\r\n\r\nContext : %\r\n\r\nData : % ' , v_err_txt ,  v_context, p_tags ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_tags(varchar[]) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_shapes(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_shapes(p_shapes_data character varying[] DEFAULT NULL::character varying[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting shapes and its correspnding edges and faces, which belong to last cache inserted, in triplets: [ext_shape_id]-,-[ext_edge_id_0,ext_edge_id_1,...]-,-[ext_face_id_0,ext_face_id_1,...]
 *		 select * from l3d.sp_insert_shapes(ARRAY['[0]-,-[4,5]-,-[]', '[1]-,-[2]-,-[0,1]']) as a   
 */	
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_shapes');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer := 0;
BEGIN 	
		
	if p_shapes_data is null then
		if v_exist_table then
			if exists(select * from tmp_ins_shapes) then					
			
				with ins_shape as
				(
					insert into l3d.shape(ext_shape_id, cache_id)
					select ext_shape_id, v_cache_id
					from 
					(
						select a.ext_shape_id from tmp_ins_shapes a order by a.ext_shape_id
					) a					
					returning id, ext_shape_id
				), ins_stats as
			    (
					update l3d.ins_stats s set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + (select count(*) from ins_shape)
					where cache_id = v_cache_id and table_name = 'shape'			    	
			    )	
				insert into tmp_insert_shape_range(min_id, max_id, offset_id)
				select min_id, max_id, offset_id
				from
				(
					select min(s.ext_shape_id) min_id, max(s.ext_shape_id) max_id, (s.id - s.ext_shape_id) offset_id
					from ins_shape s
					group by (s.id - s.ext_shape_id)
				) a
				where not exists( select * from tmp_insert_shape_range t where t.min_id = a.min_id and t.max_id = a.max_id and a.offset_id = a.offset_id)				
				;
			
				with ins_edge as
				(
					insert into l3d.edge (ext_edge_id, shape_id, cache_id)
					select a.ext_edge_id , b.offset_id + a.ext_shape_id as shape_id, v_cache_id
					from tmp_ins_edges a, tmp_insert_shape_range b 
					where a.ext_shape_id between b.min_id and b.max_id
					order by 2, 1
					returning id, ext_edge_id, shape_id						
			    ), ins_stats as
			    (
					update l3d.ins_stats s set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + (select count(*) from ins_edge)
					where cache_id = v_cache_id and table_name = 'shape'			    	
			    )
				insert into tmp_insert_edge_range(shape_id, min_id, max_id, offset_id)
				select shape_id, min_id, max_id, offset_id
				from
				(
					select s.shape_id, min(s.ext_edge_id) min_id , max(s.ext_edge_id) max_id, (s.id - s.ext_edge_id) offset_id
					from ins_edge s				 
					group by s.shape_id, (s.id - s.ext_edge_id)
				) a 
				where not exists( select * from tmp_insert_edge_range t where t.shape_id = a.shape_id and t.min_id = a.min_id and t.max_id = a.max_id and a.offset_id = a.offset_id)
				;
			
				with ins_face as
				(
					insert into l3d.face (ext_face_id, shape_id, cache_id)
					select  a.ext_face_id,  b.offset_id + a.ext_shape_id as shape_id, v_cache_id
					from tmp_ins_faces a, tmp_insert_shape_range b
					where a.ext_shape_id between b.min_id and b.max_id
					order by 2, 1
					returning id, ext_face_id, shape_id
				), ins_stats as
			    (
					update l3d.ins_stats s set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + (select count(*) from ins_face)
					where cache_id = v_cache_id and table_name = 'shape'			    	
			    )
				insert into tmp_insert_face_range(shape_id, min_id, max_id, offset_id)
				select shape_id, min_id, max_id, offset_id
				from
				(				
					select s.shape_id, min(s.ext_face_id) min_id , max(s.ext_face_id) max_id, (s.id - s.ext_face_id) offset_id
					from ins_face s
					group by s.shape_id, (s.id - s.ext_face_id)
				) a
				where not exists( select * from tmp_insert_face_range t where t.shape_id = a.shape_id and t.min_id = a.min_id and t.max_id = a.max_id and a.offset_id = a.offset_id)
				;
			
									
				truncate table tmp_ins_shapes;		
				truncate table tmp_ins_edges;			
				truncate table tmp_ins_faces;			
			
			end if;	
		end if;		
	elsif array_length(p_shapes_data, 1) > 0 then
	
		if not v_exist_table then		
			insert into l3d.ins_stats(cache_id, table_name, time_start) select ( select max(id) from tmp_insert_cache ), 'shape', current_timestamp;		

			create temp table tmp_ins_shapes(ext_shape_id integer, ext_edges_ids integer[], ext_faces_ids integer[]);
			create unique index idx_tmp_ins_shapes on tmp_ins_shapes(ext_shape_id);
		
			create temp table tmp_ins_faces(ext_face_id integer, ext_shape_id integer);		
			create index idx_tmp_ins_faces on tmp_ins_faces(ext_face_id);		
		
			create temp table tmp_ins_edges(ext_edge_id integer, ext_shape_id integer);		
			create index idx_tmp_ins_edges on tmp_ins_edges(ext_edge_id);		
		
			create temp table tmp_insert_shape_range(min_id integer, max_id integer, offset_id bigint);
			create unique index idx_tmp_insert_shape_range on tmp_insert_shape_range(min_id, max_id);
		
			create temp table tmp_insert_edge_range(shape_id bigint, min_id integer, max_id integer, offset_id bigint);
			create unique index idx_tmp_insert_edge_range on tmp_insert_edge_range(shape_id, min_id, max_id);

			create temp table tmp_insert_face_range(shape_id bigint, min_id integer, max_id integer, offset_id bigint);
			create unique index idx_tmp_insert_face_range on tmp_insert_face_range(shape_id, min_id, max_id);				
				
		end if;
	
		with sel_data as
		(
			select rd_array[1]::integer as ext_shape_id, string_to_array( rd_array[2], ',')::integer[] as ext_edges_ids, string_to_array(rd_array[3], ',')::integer[] as ext_faces_ids
			from 
			(
				select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_shapes_data ) as rd				
			) as a		
		), ins_tmp_shapes as
		(
			insert into tmp_ins_shapes(ext_shape_id) select ext_shape_id from sel_data
			returning ext_shape_id		
		), ins_tmp_edges as
		(
			insert into tmp_ins_edges(ext_edge_id, ext_shape_id) select unnest( ext_edges_ids ) as ext_edge_id, ext_shape_id from sel_data		
			returning ext_edge_id			
		), ins_tmp_faces as
		(
			insert into tmp_ins_faces(ext_face_id, ext_shape_id) select unnest( ext_faces_ids ) as ext_face_id, ext_shape_id from sel_data		
			returning ext_face_id			
		)
		select ( ( select count(*)  from ins_tmp_shapes )  + ( select count(*)  from ins_tmp_edges ) + ( select count(*)  from ins_tmp_faces ) ) ::integer into v_row_count;

		select max(ext_shape_id) into v_result from tmp_ins_shapes;	
	
		perform l3d.sp_insert_stats('shape', v_row_count );
 
	
	end if;

	return v_result;	
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_shapes Message => %\r\n\r\nContext : %\r\n\r\nData : % ' , v_err_txt ,  v_context, p_shapes_data ; 
	end ;	
end
$function$;

ALTER FUNCTION l3d.sp_insert_shapes(varchar[]) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_shape_instances(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_shape_instances(p_shape_instances_data varchar[] DEFAULT NULL::varchar[])
 RETURNS bigint
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting shapes instances and its corresponding ext shape id and ext node id, which belong to last cache inserted, in triplets: [ext_shape_instance_id]-,-[ext_shape_id]-,-[ext_node_id]
 *		 select * from l3d.sp_insert_shape_instances(ARRAY['[0]-,-[0]-,-[10]','[1]-,-[2]-,-[30]']) as a   
 */	
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_shape_instance');		
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
BEGIN 		
	if p_shape_instances_data is null then					

		if v_exist_table then 
			if exists(select * from tmp_ins_shape_instance)  then
						
				with ins_shape_instance as 
				(				
					insert into l3d.shape_instance(shape_instance_id, shape_id, node_id, cache_id) 
					select ext_shape_instance_id, shape_id, node_id, cache_id
					from 
					(
						select shi.ext_shape_instance_id, sha.offset_id + shi.ext_shape_id as shape_id, shi.ext_node_id + rng.offset_id as node_id, v_cache_id as cache_id			 
						from tmp_ins_shape_instance shi, tmp_insert_shape_range sha, tmp_insert_node_range rng
						where 
							shi.ext_shape_id between sha.min_id and sha.max_id and
							shi.ext_node_id between rng.min_id and rng.max_id					
					) a 		
					returning id, shape_id, shape_instance_id
				)
			
				insert into tmp_insert_shape_instance_range(min_id, max_id, offset_id, shape_id)
				select min_id, max_id, offset_id, shape_id
				from
				(
					select min(shape_instance_id) min_id, max(shape_instance_id) max_id, (id - shape_instance_id) offset_id, shape_id
					from ins_shape_instance
					group by shape_id, (id - shape_instance_id)
				) a
				where not exists(select * from tmp_insert_shape_instance_range t where t.min_id = a.min_id and t.max_id = a.max_id and t.offset_id = a.offset_id and t.shape_id = a.shape_id)
				;
			
				get diagnostics v_row_count = ROW_COUNT;
			
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'shape_instance';		
			
				truncate table tmp_ins_shape_instance;
			
			end if;					
		end if;
				
 	elsif array_length(p_shape_instances_data, 1) > 0 then
	
		if not v_exist_table then				
		
			create temp table tmp_ins_shape_instance(ext_shape_instance_id integer, ext_shape_id integer, ext_node_id integer);			
			create temp table tmp_insert_shape_instance(id bigint, shape_id bigint, shape_instance_id integer);					
			create index idx_tmp_insert_shape_instance on tmp_insert_shape_instance(shape_instance_id);		
		
			create temp table tmp_insert_shape_instance_range(min_id integer, max_id integer, offset_id bigint, shape_id bigint);
			create  index idx_tmp_insert_shape_instance_range on tmp_insert_shape_instance_range(shape_id, min_id, max_id);		
		
		end if;
	
		insert into tmp_ins_shape_instance
		select rd_array[1]::integer as ext_shape_instance_id, rd_array[2]::integer as ext_shape_id, substring( rd_array[3], '[0-9]+' )::integer as ext_node_id
		from
		(				
			select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_shape_instances_data ) as rd		
		) as a;	
	
		perform l3d.sp_insert_stats('shape_instance', array_length(p_shape_instances_data, 1));	
		
	end if;
		
	return 0;
end
$function$;

ALTER FUNCTION l3d.sp_insert_shape_instances(varchar[]) OWNER TO i3dhub;


drop function if exists l3d.sp_insert_face_edge_instances(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_face_edge_instances(p_face_edge_instances_data varchar[] DEFAULT NULL::varchar[])
 RETURNS bigint
 LANGUAGE plpgsql
AS $function$
declare
/**
*   Usage:
*        Inserting face/edge instances, which belong to last cache inserted, in triplets: [e/f]-,-[ext_edge_id/ext_face_id]-,-[ext_shape_instance_id]-,-[ext_original_face_id]
*        select * from l3d.sp_insert_face_edge_instances(ARRAY['[e]-,-[1]-,-[0]','[f]-,-[2]-,-[1]-,-[2015]']) as a  
 */   
   v_exist_table bool := sp_exists_temporary_table('tmp_ins_face_edge_instance');   
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
begin
		
	if p_face_edge_instances_data is null  then
	
          if v_exist_table then 
          	if exists(select * from tmp_ins_face_edge_instance) then	

		         with ins_edge_instance as 
		      	(
					insert into l3d.edge_instance(edge_id, shape_instance_id, cache_id) 
					select e.offset_id + ei.ext_face_edge_id as edge_id, si.offset_id + ei.ext_shape_instance_id, v_cache_id
					from tmp_ins_face_edge_instance ei, tmp_insert_edge_range e, tmp_insert_shape_instance_range si
					where 
						ei.face_or_edge = 'e' 
					and ei.ext_face_edge_id between e.min_id and e.max_id 					
					and ei.ext_shape_instance_id between si.min_id and si.max_id
					and si.shape_id = e.shape_id
					returning id, edge_id, shape_instance_id 
				)
				insert into tmp_insert_edge_instance select * from ins_edge_instance ;          	
			
				get diagnostics v_row_count = ROW_COUNT;
			
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count
				where cache_id = v_cache_id and table_name = 'face_edge_instance';		
							
							
				with join_face_instance as
				(
					select f.offset_id + fi.ext_face_edge_id as face_id, si.offset_id + fi.ext_shape_instance_id as shape_instance_id, v_cache_id as cache_id, ext_original_face_id
					from tmp_ins_face_edge_instance fi, tmp_insert_face_range f, tmp_insert_shape_instance_range si
					where 
						fi.face_or_edge ='f'  
					and fi.ext_face_edge_id between f.min_id and f.max_id
--					and si.shape_instance_id = fi.ext_shape_instance_id
					and fi.ext_shape_instance_id between si.min_id and si.max_id
					and si.shape_id = f.shape_id
				),		
				ins_face_instance as
				(
					insert into l3d.face_instance (face_id, shape_instance_id, cache_id) 
					select face_id, shape_instance_id, cache_id 
					from join_face_instance			
					returning id, face_id, shape_instance_id
				), ins_face_original as
				(
					insert into l3d.face_instance_to_original(face_instance_id, original_face_id)
					select  fi.id, jfi.ext_original_face_id
					from join_face_instance jfi , ins_face_instance fi
					where  
						jfi.face_id = fi.face_id and jfi.shape_instance_id = fi.shape_instance_id and jfi.ext_original_face_id is not null							
				)			
				insert into tmp_insert_face_instance select * from ins_face_instance;
							
				get diagnostics v_row_count = ROW_COUNT;
			
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'face_edge_instance';		
												
				truncate table tmp_ins_face_edge_instance;
			
	          end if;		          
          end if;
	
	elsif array_length(p_face_edge_instances_data, 1) > 0 then
  
		if not v_exist_table then
		
                 create temp table tmp_insert_face_instance(id bigint, face_id bigint, shape_instance_id bigint);                  
                 create temp table tmp_insert_edge_instance(id bigint, edge_id bigint, shape_instance_id bigint);                                                                     
                
				create temp table tmp_ins_face_edge_instance(face_or_edge varchar(1), ext_face_edge_id integer, ext_shape_instance_id integer, ext_original_face_id integer) ;               
                
		end if;
                  
		insert into tmp_ins_face_edge_instance
		select rd_array[1] as face_or_edge, rd_array[2]::integer as ext_face_edge_id, substring( rd_array[3], '[0-9]+' )::integer as ext_shape_instance_id, substring( rd_array[4], '[0-9]+' )::integer as ext_original_face_id
		from
		(			
			select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_face_edge_instances_data ) as rd			
		) as r;          	
	
		perform l3d.sp_insert_stats('face_edge_instance', array_length(p_face_edge_instances_data, 1));
	                          
   end if;
 
	return 0;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_face_edge_instances Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_face_edge_instances_data ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_face_edge_instances(varchar[]) OWNER TO i3dhub;

drop FUNCTION if exists l3d.sp_insert_aux_nodes(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_aux_nodes(p_aux_nodes_data varchar[] DEFAULT NULL::varchar[])
 RETURNS bigint
 LANGUAGE plpgsql
AS $function$
declare
/**
*   Usage:
*        Inserting aux Nodes which belong to last cache inserted, in quadruplets : [ext_aux_node_id]-,-[type]-,-[label],--[aux_node_facelink_1,aux_node_facelink_2...]
*              select * from  l3d.sp_insert_aux_nodes(ARRAY['[0]-,-[I3DH Internal Grouping]-,-[PMI]','[1]-,-[I3DH Internal Grouping]-,-[CoordinateSystem]-,-[$NGID<chain>="__PLM_INST_TH_UID"\0QGD3DM0000040 1\0QGD3DM0000041 \0QGD3DM0000041_1_1\0\0/6714]','[2]-,-[Special Marker that doesn''t delete]-,-[Axis System.1]']);  
 * select * from l3d.sp_insert_aux_nodes(ARRAY[[0]-,-[I3DH Internal Grouping]-,-[PMI],[1]-,-[I3DH Internal Grouping]-,-[CoordinateSystem],[2]-,-[Special Marker]-,-[Axis System.1]]) as a
*   
 */
   v_exist_table bool := sp_exists_temporary_table('tmp_ins_aux_node');	  
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result bigint := 0;
begin
	
		
	if p_aux_nodes_data is null then
  
		if v_exist_table then
		
			if exists(select * from tmp_ins_aux_node) then			 
										
				with ins_aux_node as
				(
					insert into l3d.aux_node(aux_node_type_id, cache_id, ext_aux_node_id, label)			
					select type_id, v_cache_id as cache_id, ext_aux_node_id, label
					from
					(
						select type_id, nod.ext_aux_node_id, nod.label from tmp_ins_aux_node nod order by nod.ext_aux_node_id					
					) a					
					returning id , ext_aux_node_id	         	
		         ), upd_stats as 
				(
					update l3d.ins_stats s set time_end = current_timestamp, records_commited = records_commited  + (select count(*) from ins_aux_node)
					where cache_id = v_cache_id and table_name = 'aux_node'						
				)
				insert into tmp_insert_aux_node_range(min_id, max_id, offset_id)
				select min(ext_aux_node_id), max(ext_aux_node_id), (id - ext_aux_node_id)
				from ins_aux_node i
				group by (id - ext_aux_node_id);
			
				insert into l3d.foreign_face_link(face_link, aux_node_id)
				select unnest(dat.foreign_faces_links) as face_link, rng.offset_id + dat.ext_aux_node_id
				from tmp_ins_aux_node dat, tmp_insert_aux_node_range rng
				where dat.foreign_faces_links is not null and  dat.ext_aux_node_id between rng.min_id and rng.max_id;
									
				truncate table tmp_ins_aux_node;
			
			end if;			
		
		end if;	                                 
   elsif array_length(p_aux_nodes_data, 1) > 0 then
  
		if not v_exist_table then
				
			create temp table tmp_ins_aux_node(ext_aux_node_id integer, type_name varchar, label varchar, foreign_faces_links varchar[],type_id integer);
			create unique index idx_tmp_ins_aux_node on tmp_ins_aux_node(ext_aux_node_id);

			create index idx_tmp_ins_aux_node_ffl on tmp_ins_aux_node using hash(foreign_faces_links);		
								
			create temp table tmp_insert_aux_node_range(min_id integer, max_id integer, offset_id bigint);		
			create unique index idx_tmp_insert_aux_node_range on tmp_insert_aux_node_range(min_id,max_id);		
															
		end if;
			
		insert into tmp_ins_aux_node (ext_aux_node_id, type_name, label, foreign_faces_links, type_id)		 		 
		select ext_aux_node_id, type_name, label, foreign_faces_links, typ.id
		from
		(
			select rd_array[1]::integer as ext_aux_node_id, rd_array[2] as type_name, replace( rd_array[3], '\"', '"' )  as label, string_to_array( rd_array[4], ',' ) as foreign_faces_links
			from
			(			
				select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_aux_nodes_data ) as rd				
			) as a
		) a
		  left join l3d.aux_node_type typ on typ.name = a.type_name;		 
		 
		if exists(select * from tmp_ins_aux_node where type_id is null) then 
		 
			with ins_aux_node_type as 
			(
				 insert into l3d.aux_node_type(name) select type_name from tmp_ins_aux_node where type_id is null group by type_name
				returning id, name
			)
			update tmp_ins_aux_node ins set type_id = id from ins_aux_node_type typ where ins.type_name = typ.name;
		 
		end if;
	
		select max(ext_aux_node_id) into v_result from tmp_ins_aux_node;
	
		perform l3d.sp_insert_stats('aux_node', array_length(p_aux_nodes_data, 1));	
				 		 					
   end if;
 
   return v_result;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_aux_nodes Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_aux_nodes_data ; 
	end ;  
end
$function$;

ALTER FUNCTION l3d.sp_insert_aux_nodes(varchar[]) OWNER TO i3dhub;

drop FUNCTION if exists l3d.sp_insert_aux_nodes_attributes(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_aux_nodes_attributes(p_aux_attributes_values varchar[] DEFAULT NULL::varchar[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage: 
 *   // Inserting aux attributes, which their keys and values are given, where their aux node was the last one inserted   : [aux_node_id]-,-[key]-,-[value]
 *    select * from l3d.sp_insert_aux_nodes_attributes(ARRAY['[0]-,-[CLIP_POSITION]-,-[-65.739510,-770.471724,346.265323]','[1]-,-[PMITextForegroundColor]-,-[0x0000f000]','[2]-,-[Text]-,-[Werkstattbeschnitt]']) as a
 */	
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_aux_attributes');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
begin
	if p_aux_attributes_values is null then
		if v_exist_table then 
			if exists(select * from tmp_ins_aux_attributes)  then
			
				insert into l3d.aux_attributes (aux_node_id, key, value) 
				select a.ext_aux_node_id + rng.offset_id, a.key, a.value
				from 
					(
						select key, value, unnest(nodes_ids) as ext_aux_node_id 
						from tmp_ins_aux_attributes
					) a,  tmp_insert_aux_node_range rng
				where
					a.ext_aux_node_id between rng.min_id and rng.max_id;			
									
				select count(*) into v_row_count from tmp_ins_aux_attributes;
			
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'aux_attributes';		
					
				truncate table tmp_ins_aux_attributes;

			end if;		
		end if;
	elsif array_length(p_aux_attributes_values, 1) > 0 then

		if not v_exist_table then		
			create temp table tmp_ins_aux_attributes(key varchar, value varchar, nodes_ids integer[]);								
		end if;
	
	
	
		insert into tmp_ins_aux_attributes(key, value, nodes_ids)
		select a.key, a.value, array_agg(ext_aux_node_id) 
		from
		(
			select rd_array[1]::integer as ext_aux_node_id,
					rd_array[2] as key,
					replace( rd_array[3], '\"', '"' )  as value
			from
			(						
				select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_aux_attributes_values ) as rd				
			) as a			
		) a
		group by a.key, a.value;		

		get diagnostics v_row_count = ROW_COUNT;	
	
		perform l3d.sp_insert_stats('aux_attributes', v_row_count);
	
											
	end if;

	return v_row_count;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_aux_nodes_attributes Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_aux_attributes_values ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_aux_nodes_attributes(varchar[]) OWNER TO i3dhub;

drop FUNCTION if exists l3d.sp_insert_aux_nodes_values(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_aux_nodes_values(p_aux_nodes_values character varying[] DEFAULT NULL::character varying[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting aux Nodes values which belong to last cache inserted, in quadruplets : [aux_ext_node_id]-,-[value type (s/n)]-,-[value name]-,-[value]
 *		 select * from l3d.sp_insert_aux_nodes_values(ARRAY['[0]-,-[s]-,-[label]-,-[#Notes and Remarks]','[1]-,-[s]-,-[markerType]-,-[CoordinateSystem]','[1]-,-[n]-,-[value]-,-[0.322886]']) as a   
 */		
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_aux_value');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
BEGIN 	
	if  p_aux_nodes_values is null then	
		if v_exist_table then 
			if exists(select * from tmp_ins_aux_value) then
										
				with ins_name as 
				(
					insert into l3d.aux_value_name(name)  select value_name from tmp_ins_aux_value where value_name_id is null group by value_name
					returning id, name
				) , ins_tmp as 
				(
					insert into tmp_ins_aux_name(id, name) select id, name from ins_name
					returning id, name
				)			
				insert into l3d.aux_value(aux_value_name_id, value_str, value_num, aux_node_id)			
				select a.value_name_id, 
						case when a.value_type = 's' then replace(a.value, '\"', '"' ) end, 
						case when a.value_type = 'n' then replace(a.value, ',', '.')::numeric end,
						a.ext_node_id + rng.offset_id
				from 
				(
					select value_type, value, unnest( nodes_ids ) as ext_node_id, coalesce( s.value_name_id, n.id ) as  value_name_id 
					from tmp_ins_aux_value s
					left join ins_tmp n on s.value_name_id is null and s.value_name = n.name					
				) a, tmp_insert_aux_node_range rng
				where  a.ext_node_id between rng.min_id and rng.max_id;
																					
			 	select count(*) into v_row_count from tmp_ins_aux_value;			
						
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'aux_value';		
					
				truncate table tmp_ins_aux_value;

			
			end if;						
		end if;	
	elsif array_length(p_aux_nodes_values, 1) > 0 then
	
		if not v_exist_table then 
				
			create temp table tmp_ins_aux_value(value_name varchar, value_type varchar, value varchar, nodes_ids integer[], value_name_id integer);
		
			create index idx_tmp_ins_aux_value_name_id on tmp_ins_aux_value(value_name_id);		
		
		
			create temp table tmp_ins_aux_name(id integer, name varchar);
			create index idx_tmp_ins_aux_name on tmp_ins_aux_name using hash(name);
		
			insert into tmp_ins_aux_name(id, name) select id, name from l3d.aux_value_name ; 			
		
		
		end if;
	
		insert into tmp_ins_aux_value(value_name, value_type, value, nodes_ids, value_name_id)
		select  value_name, value_type, value, nodes_ids, vn.id
		from 
		(
			select	value_name, value, array_agg( ext_aux_node_id )  as nodes_ids, value_type
			from
			(
				select rd_array[1]::integer as ext_aux_node_id, rd_array[2] as value_type, rd_array[3] as value_name, rd_array[4] as value
				from
				(				
					select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_aux_nodes_values ) as rd					
				) as a		
			) as a						
			group by value_name, value, value_type
		) a				
		 left join tmp_ins_aux_name vn on vn.name = a.value_name;
	--	left join aux_value_name vn on vn.name = a.value_name;
	
		get diagnostics v_row_count = ROW_COUNT;	

		perform l3d.sp_insert_stats('aux_value', v_row_count);	
									
	end if;

	return  v_row_count;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_aux_nodes_values Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_aux_nodes_values ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_aux_nodes_values(varchar[]) OWNER TO i3dhub;

drop FUNCTION if exists l3d.sp_insert_aux_nodes_relations(varchar[]);

CREATE OR REPLACE FUNCTION l3d.sp_insert_aux_nodes_relations(p_aux_nodes_relations varchar[] DEFAULT NULL::varchar[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting aux Nodes relations which belong to last cache inserted, in duplets : [pred_ext_aux_node_id]-,-[succ_ext_aux_node_id_1, succ_ext_aux_node_id_2...]
 *		 select * from l3d.sp_insert_aux_nodes_relations(ARRAY['[0]-,-[1,2]','[1]-,-[2]']) as a   
 */		
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_aux_node_to_aux_node');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
BEGIN 		
	
	if p_aux_nodes_relations is null then 
      if v_exist_table then 
          	if exists(select * from tmp_ins_aux_node_to_aux_node) then	
									
				insert into l3d.aux_node_to_aux_node(pred_id, succ_id)
				select nd_pred.offset_id + a.ext_node_pred_id, nd_succ.offset_id + a.ext_node_succ_id
				from 
					(select ext_node_pred_id, unnest(ext_node_succ_ids) as ext_node_succ_id from tmp_ins_aux_node_to_aux_node) a 
					, tmp_insert_aux_node_range  nd_succ, tmp_insert_aux_node_range  nd_pred 
				where
						a.ext_node_succ_id between nd_succ.min_id and nd_succ.max_id and
						a.ext_node_pred_id between nd_pred.min_id and nd_pred.max_id;									
          	
						
--				get diagnostics v_row_count = ROW_COUNT;
				select count(*) into v_row_count from tmp_ins_aux_node_to_aux_node;
								
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'aux_node_to_aux_node';		
			
				truncate table tmp_ins_aux_node_to_aux_node;
          	
	    	end if;	
		end if;	    
	elsif array_length(p_aux_nodes_relations, 1) > 0 then
	
      	if not v_exist_table then
	      		
			create temp table tmp_ins_aux_node_to_aux_node(ext_node_pred_id integer, ext_node_succ_ids integer[]);
      	
		end if;      	
	
		insert into tmp_ins_aux_node_to_aux_node (ext_node_pred_id, ext_node_succ_ids)
		select rd_array[1]::integer as pred_id, ( string_to_array( rd_array[2], ',' ) )::integer[]  as succ_ids
		from
		(					
			select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array  from unnest( p_aux_nodes_relations ) as rd			
		) as a;
	
		get diagnostics v_row_count = ROW_COUNT;		
		
		perform l3d.sp_insert_stats('aux_node_to_aux_node', v_row_count);	
					
	end if;
		
	return v_row_count;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_aux_nodes_relations Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_aux_nodes_relations ; 
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_aux_nodes_relations(varchar[]) OWNER TO i3dhub;

drop function if exists l3d.sp_insert_aux_node_topo_relations(varchar[]); 

CREATE OR REPLACE FUNCTION l3d.sp_insert_aux_node_topo_relations(p_topo_relations_data varchar[] DEFAULT NULL::varchar[])
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare
/**
 *   Usage:
 *		Inserting topo relations, which belong to last cache inserted, in quadruplets: [e/f]-,-[ext_edge_id/ext_face_id]-,-[ext_shape_instance_id]-,-[ext_node_id]
 *		 select * from l3d.sp_insert_aux_node_topo_relations(ARRAY['[e]-,-[1]-,-[0]-,-[0]','[f]-,-[2]-,-[1]-,-[0]']) as a   
 */		
	v_exist_table bool := sp_exists_temporary_table('tmp_ins_topo_relations');
	v_cache_id integer := ( select max(id) from tmp_insert_cache );
	v_row_count integer;
	v_result integer;
begin

	
	if p_topo_relations_data is null then
	
		if v_exist_table then
		
			if exists(select * from tmp_ins_topo_relations_data)  then			
							
				with sel_relations as
				(
					select	rd_array[1] as edge_or_face, rd_array[2]::integer as ext_edge_face_id, rd_array[3]::integer as ext_shape_instance_id, rd_array[4]::integer as ext_node_id
					from
					( 
						select string_to_array( right(left( rd, -1), -1), ']-,-[' ) as rd_array 											
						from
						(
							 select unnest(relations) as rd from tmp_ins_topo_relations_data			
						) a
						
					) as a				
				), ins_edge_intance_to_aux_node as
				(
					insert into edge_instance_to_aux_node(aux_node_id, edge_instance_id)
					select a.ext_node_id + rng.offset_id, ei.id
					from 
					sel_relations a, tmp_insert_aux_node_range rng, tmp_insert_edge_range e, tmp_insert_shape_instance_range si, tmp_insert_edge_instance ei
					where
						a.edge_or_face = 'e' 
					and a.ext_node_id between rng.min_id and rng.max_id  
					and a.ext_edge_face_id  between e.min_id and e.max_id				 
	
					and a.ext_shape_instance_id between si.min_id and si.max_id				  
					and si.shape_id = e.shape_id
					
					and ei.edge_id = a.ext_edge_face_id + e.offset_id 
					and ei.shape_instance_id = a.ext_shape_instance_id + si.offset_id 								
					returning aux_node_id, edge_instance_id
				), upd_stats as
				(
					update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + (select count(*) from ins_edge_intance_to_aux_node) where cache_id = v_cache_id and table_name = 'topo_relations'				
				)
				insert into l3d.face_instance_to_aux_node(aux_node_id, face_instance_id)
				select a.ext_node_id + rng.offset_id, fi.id 
				from sel_relations a , tmp_insert_aux_node_range rng, tmp_insert_face_range f, tmp_insert_shape_instance_range si, tmp_insert_face_instance fi
				where 
					a.edge_or_face = 'f'			
				and a.ext_node_id between rng.min_id and rng.max_id 
				and a.ext_edge_face_id  between f.min_id and f.max_id 
				
				and a.ext_shape_instance_id between si.min_id and si.max_id  
				and si.shape_id = f.shape_id				
				
				and fi.face_id = a.ext_edge_face_id + f.offset_id 
				and fi.shape_instance_id = a.ext_shape_instance_id + si.offset_id 				
				;			

				get diagnostics v_row_count = ROW_COUNT;
									
				update l3d.ins_stats set time_end = current_timestamp, records_commited = coalesce( records_commited, 0 )  + v_row_count where cache_id = v_cache_id and table_name = 'topo_relations';					
			
				truncate table tmp_ins_topo_relations;
				truncate table tmp_ins_topo_relations_data;
			
			end if;
		
		end if;
	
	elsif array_length(p_topo_relations_data, 1) > 0 then
				
		if not v_exist_table then 		
			
			create temp table tmp_ins_topo_relations(edge_or_face varchar(1), ext_edge_face_id integer, ext_shape_instance_id integer,ext_node_id integer);		
		
			create temp table tmp_ins_topo_relations_data(relations varchar[]);		
		
		end if;
	
		insert into tmp_ins_topo_relations_data select p_topo_relations_data;
		
		perform l3d.sp_insert_stats('topo_relations', array_length(p_topo_relations_data, 1));
							
	end if;
				
	return 0;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_aux_node_topo_relations Message => %, Context : % Data : % ' , v_err_txt ,  v_context, p_topo_relations_data ;
	end ;
end
$function$;

ALTER FUNCTION l3d.sp_insert_aux_node_topo_relations(varchar[]) OWNER TO i3dhub; 


-- FUNCTION : l3d.sp_insert_statistics(json default null);

drop function if exists l3d.sp_insert_statistics(json);

CREATE FUNCTION sp_insert_statistics(p_stats json DEFAULT NULL::json) RETURNS integer
    LANGUAGE plpgsql
    AS $function$
declare
/**
 *   Usage:
 *		Inserting aux Nodes relations which belong to last cache inserted, in duplets : [succ_ext_aux_node_id]-,-[pred_ext_aux_node_id_1, pred_ext_aux_node_id_2...]
 *		 select * from l3d.sp_insert_aux_nodes_relations(ARRAY['[0]-,-[1,2]','[1]-,-[2]']) as a   
 */
	initial_db_id int4 := null;
	n_cache_id integer;
	v_exist_table bool := sp_exists_temporary_table('tmp_insert_data_stats');
begin
		
	if p_stats is null then
	
		if v_exist_table then
	
			if not exists( select * from data_stats where cache_id in (select cache_id from tmp_insert_data_stats) ) then
				insert into l3d.data_stats select * from tmp_insert_data_stats; 				
			end if;
			
			truncate table tmp_insert_data_stats;	
		end if;	
		
	else
		if not v_exist_table then	
			create temp table tmp_insert_data_stats as select * from l3d.data_stats where false;	 
		end if;
	

		if not exists( select * from tmp_insert_data_stats) then 
	
			select max(id) into n_cache_id from tmp_insert_cache;
			
			insert into tmp_insert_data_stats(cache_id, stats) values(n_cache_id, p_stats);
		
		end if;
	
	end if;
		
	return 0;

end 
$function$;

ALTER FUNCTION l3d.sp_insert_statistics(json) OWNER TO i3dhub;

-- function: l3d.sp_insert_clear();

drop function if exists l3d.sp_insert_clear();

CREATE OR REPLACE FUNCTION l3d.sp_insert_clear()
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
begin 
	
		if sp_exists_temporary_table('tmp_insert_node_range') then								
			truncate tmp_insert_node_range;
		end if;
	
		if sp_exists_temporary_table('tmp_insert_aux_node_range') then		
			truncate tmp_insert_aux_node_range;		
		end if;
	
		if sp_exists_temporary_table('tmp_insert_shape_range') then				
				truncate tmp_insert_shape_range;			
		end if;
	
		if sp_exists_temporary_table('tmp_insert_edge_range') then			
				truncate tmp_insert_edge_range;			
		end if;
	
		if sp_exists_temporary_table('tmp_insert_face_range') then			
				truncate tmp_insert_face_range;			
		end if;
	
		if sp_exists_temporary_table('tmp_insert_shape_instance_range') then			
--				truncate tmp_insert_shape_instance;					
				truncate tmp_insert_shape_instance_range;			
		end if;
	
		if sp_exists_temporary_table('tmp_insert_edge_instance') then	
				truncate tmp_insert_edge_instance;					
		end if;
	
		if sp_exists_temporary_table('tmp_insert_face_instance') then	
				truncate tmp_insert_face_instance;						
		end if;	


	
	return 0;
end
$function$;	

ALTER FUNCTION l3d.sp_insert_clear() OWNER TO i3dhub;

-- function: l3d.sp_insert_flush();

drop function if exists l3d.sp_insert_flush(bool);

CREATE OR REPLACE FUNCTION l3d.sp_insert_flush(p_clear bool default true)
 RETURNS integer
 LANGUAGE plpgsql
AS $function$
declare     
/*   Usage:
 *    // Flushing data from temporary tables to tables
 *    select * from l3d.sp_insert_flush() as a
 */
	v_result integer := 0;
BEGIN 		
		
	perform sp_insert_nodes();

	perform sp_insert_metadata();	

	perform sp_insert_tags();

	perform sp_insert_shapes();
							
	perform sp_insert_shape_instances();

	perform sp_insert_face_edge_instances();
			
	perform sp_insert_aux_nodes();	

	perform sp_insert_aux_nodes_values();
	
	perform sp_insert_aux_nodes_attributes();
	
	perform sp_insert_aux_nodes_relations();

	perform sp_insert_aux_node_topo_relations();
			
	perform sp_insert_statistics();

--	DISCARD TEMP; -- removing all previous temporary tables in case they were not committed before

	if p_clear then
		perform l3d.sp_insert_clear();		
	end if;
	

	return v_result;
exception when others then
	declare
		v_err_txt text;
		v_context text;
	begin		
--			DISCARD temp;		
		 GET STACKED DIAGNOSTICS v_err_txt = MESSAGE_TEXT, v_context = pg_exception_context; 		
		raise exception 'Error sp_insert_flush Message => %, Context : %' , v_err_txt ,  v_context ; 
	end ;				
end
$function$;


ALTER FUNCTION l3d.sp_insert_flush(bool) OWNER TO i3dhub;

